/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/



/*
  Constructs a BoxPlacement instance for all CapoBlocks.
  This instance consists of:

  1) ConstrainingHGraph over all blocks (HGraph + 4 nodes/box + extra edges)
	see constrainingHG.h
  2) initial placement:  Each core-cell node is in the middle of it's
	CapoBlock, the constraining nodes are at the corners of their
	CapoBlocks.	
  3) fixed constraints.  All pads, placed cells and the 4*#blocks constraining
	nodes are fixed
*/

#ifndef __CAPO_BOXPL_H__
#define __CAPO_BOXPL_H__

#include "constrainingHG.h"
#include "capoBlock.h"
#include "BoxPlace/boxplace.h"

class BBoxPlacer;

class BoxPlacementForCapo
{
    const vector<CapoBlock*>&		_allBlocks;
    const HGraphFixed&			_netlistHG;

    BBoxPlacer*				_boxPl;

  public:
    BoxPlacementForCapo(
	const vector<CapoBlock*>&	allBlocks,
        const HGraphFixed&		hgraphOfNetlist,
	const Placement&		origPl,
	const BBoxPlacer::Parameters&   params);
	
    ~BoxPlacementForCapo();

    void setPlacement(Placement& plToSet);

    bool placementIsValid();
};

#endif
