/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




// Created: Sep 11, 1997 by Igor Markov & Andy Caldwell
// Portions based on Wenyong Deng's Cadence-confidential outline.

//CHANGES
//970923 aec moving to new Capo structure: major changes include removal pf
//		padBlock, addition of a persistent slice, move away from
//		the 'changeNetlist' 'changeLayout' design into a unified
//		'refineBlock' structure.
//971113 aec added class CapoPlacer::Parameters
//971118 aec added SA multi-start and multi-level params to 
//			CapoPlacer::Paremeters
//980215 aec moved parameter code to capoParams.cxx
//980305 ilm removed fixedConstraint handling for Partitioner::Parameters
//980401 ilm split into refineBlock.cxx, splitBlock.cxx and the rest
//990202 aec capoPlacer now operates on an RBPlacement and an HGraphWDims,
//		rather than DB
//990319 ilm the main Capo loop now copies all placed blocks to the new
//           level and stops when there are no blocks to refine

#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include <algorithm>
#include "ABKCommon/abkcommon.h"
#include "capoPlacer.h"
#include "capoBlock.h"
#include "PlaceEvals/placeEvals.h"
#include "Stats/trivStats.h"
#include "Placement/xyDraw.h"
#include "RBPlace/RBPlacement.h"
//#include "boxPlProb.h"
//#include "BoxPlace/boxplace.h"

CapoPlacer::CapoPlacer(RBPlacement& rbplace, const CapoParameters& params)
//                     const vector<char*>* altCellNames)
	: _params(params), _rbplace(rbplace), 
          _hgraphOfNetlist(rbplace.getHGraph()),
	  _coreBBox(/*BBoxFromRBPlace*/), 
	  _placement(static_cast<PlacementWOrient>(rbplace)), 
	  _kPlacement(static_cast<PlacementWOrient>(rbplace)), 
          _placed(rbplace.getFixed()),
          _pinLocCalc(_hgraphOfNetlist, _placement),
	  _layerNum(0), 
	  _cellToBlockMap(_hgraphOfNetlist.getNumNodes(), 
			  static_cast<const CapoBlock*>(NULL)),
	  _nodeSeen(_hgraphOfNetlist.getNumNodes()), 
	  _edgeSeen(_hgraphOfNetlist.getNumEdges()),
	  _numOrigNets(100,0), _numEssentialNets(100,0),
	  _origNetPins(100,0.0), _essentialNetPins(100,0.0),
	  _numProblemsOfSize(100,0),
	  _netHasBeenCut(_hgraphOfNetlist.getNumEdges(), false),
	  _numNotSolved(0), _numSmallPartProbs(0)
{
    Timer  capoTimer;

    if(_params.verb.forActions > 0)
	cerr<<endl<<"Launching UCLA's CapoPlacer"<<endl;

    setupAndCheck();

    unsigned maxBlocks = _params.stopAtBlocks ? _params.stopAtBlocks : UINT_MAX;

    bool haveUnplacedBlocks = true;
    bool doVCuts = true;


//cout<<"Cell Names "<<endl;
////for(itHGFNodeGlobal nIt = _hgraphOfNetlist.nodesBegin(); nIt != _hgraphOfNetlist.nodesEnd(); nIt++)
//	cout<<(*nIt)->getName()<<endl;


    while(haveUnplacedBlocks && _curLayout->size() < maxBlocks )
    {
	AllowedPartDir partDir;
	if(_layerNum < 3) 
	     partDir = HAndV;
	else if(_layerNum == 3 || doVCuts)
	{
	     partDir = VOnly;
	     doVCuts = false;	//next time will be horizontal
	}
	else 
	{
	     partDir = HOnly;
	     doVCuts = true;
	}

  	if(_params.useKPl || _params.propFromPl)
	{
	  _kPlacement = _placement;
	  unsigned numCellsPerBlock = _hgraphOfNetlist.getNumNodes() / 
			(_curLayout->size() + _placedBlocks.size());
  	  if(_layerNum > 1 && numCellsPerBlock > 100)
	  {
            abkfatal(0,"BoxPlacer is not available in this distribution\n");
/*
	    BBoxPlacer::Parameters bpParams(_params.verb);
	    bpParams.noY = true;
	    bpParams.maxIterations = 100;
	  
	    BoxPlacementForCapo placeAllBlocks(*_curLayout, 
		_hgraphOfNetlist, _placement, bpParams);

	    abkwarn(placeAllBlocks.placementIsValid(),
		"BoxPl placement does not respect block boundaries");

	    placeAllBlocks.setPlacement(_kPlacement);
*/

	    char plName[200];
	    sprintf(plName, "BBPlForLvl%d_%dBlocks", 
				_layerNum, _curLayout->size());
	    ofstream plSaveFile(plName);
	    plSaveFile<<_kPlacement<<endl;
	    plSaveFile.close();

            BBoxPlEval            hpWLEval   (_hgraphOfNetlist, _kPlacement);
 	    double                bboxWL  = hpWLEval.getCost();
            cout<<"Layer "<<_layerNum<<" After BoxPL:  WL   (HP): "
			<<bboxWL <<"\n";
            cout<<endl;

//	    if(_params.propFromPl)
//     placeAllBlocks.setPlacement(_placement);
	  }
	}

	haveUnplacedBlocks  = doOneLayer(partDir);

	//if replaceSmallBlocks == AtEveryLayer, the placed blocks
	//will still be in curLayout.  Otherwise, they will have
	//been moved into _placedBlocks

        if(_params.verb.forMajStats > 1)
        {
            cout << "\nTotal cut nets: " << getTotalNetCut() << endl;
            vector<unsigned> externalCutNets, containedNets;
            getNetCutInfo(externalCutNets, containedNets);

	    if(externalCutNets.size() == _curLayout->size()) 
	    {
	      cout<<"Per Block Stats:  "
		  <<"<numCells><cellArea><cutEdges>"<<endl;
	
              for(unsigned ijk = 0; ijk < externalCutNets.size(); ijk++)
	      {
                cout<<"Block " << ijk <<" has "
		    <<(*_curLayout)[ijk]->getNumCells()<<" cells, "
		    <<(*_curLayout)[ijk]->getTotalCellArea()
		    <<" cell area, "
		    <<externalCutNets[ijk] << " cut nets and "
                    <<containedNets[ijk] << " contained nets\n";
	      }
	    }
        }
    }

    if(_params.saveBlocks)
    {
           cout << "Saving block membership in block_mem.dat" << endl;
           printBlockMembership("block_mem.dat");
           cout << "Saving hierarchical cell names cell_names.dat" << endl;
           printHierCellNames("cell_names.dat");
    }


    if(_params.replaceSmallBlocks == AtTheEnd)
    {
	Timer replaceTimer;
	cout<<" Replacing Small Blocks"<<endl;
   	replaceSmallBlocks();
	replaceTimer.stop();
	cout<<" Replacing Small Blocks took: "<<
		replaceTimer.getUserTime()<<endl;
    }

    capoTimer.stop();

    if(_params.verb.forMajStats > 0)
        cout << "   CapoPlacer took : " << capoTimer << endl;

    if(_params.verb.forMajStats > 0)
    {
	cout<<"   Final Average Relative Overfill: "<<
		100*(_totalOverfill/_totalLayoutArea)<<"%"<<endl;
	cout<<"  Max 1block Overfill: "<<100*_maxOverfill<<"%"<<endl;
    }

    rbplace.resetPlacementWOri(_placement);
}


double CapoPlacer::estimateWL()
{
 /*
    LayoutBBoxes 	regions; //CapoBlocks + Points of all fixed nodes

    unsigned            numNodes = _hgraphOfNetlist.getNumNodes();

    vector<unsigned>    assignment (numNodes, UINT_MAX);
	//for each node, this gives the index into the regions vector
	//of the region it is in

    itHGFNodeGlobal node; 
    for(node = _hgraphOfNetlist.nodesBegin(); 
        node != _hgraphOfNetlist.nodesEnd(); node++)
    {
	unsigned nodeId = (*node)->getIndex();

	if(_placed[nodeId]) 
	{
	    assignment[nodeId] = regions.size();
	    regions.push_back(BBox(_placement[nodeId]));
	}
    }

    for(unsigned b = 0; b < _curLayout->size(); b++)
    {
	const CapoBlock& curBlk = *(*_curLayout)[b];

	vector<unsigned>::const_iterator nItr;
	for(nItr = curBlk.cellIdsBegin(); nItr != curBlk.cellIdsEnd(); nItr++)
	    assignment[*nItr] = regions.size();

	regions.push_back(curBlk.getBBox());
    }
 */

    Timer bboxTM;
    BBoxPlEval       bboxEval    (_hgraphOfNetlist, _placement);
    bboxTM.stop();

/*
    Timer bboxWChengTM;
    BBoxWChengPlEval bboxWCEval  (_hgraphOfNetlist, _placement);
    bboxWChengTM.stop();
    Timer bboxWRSMTTM;
    BBoxWRSMTPlEval  bboxWREval  (_hgraphOfNetlist,  _placement);
    bboxWRSMTTM.stop();

    Timer hbboxTM;
    HBBoxPlEval       hbboxEval   (_hgraphOfNetlist, regions, assignment);
    hbboxTM.stop();
    Timer hbboxWChengTM;
    HBBoxWChengPlEval hbboxWCEval (_hgraphOfNetlist, regions, assignment);
    hbboxWChengTM.stop();
    Timer hbboxWRSMTTM;
    HBBoxWRSMTPlEval  hbboxWREval (_hgraphOfNetlist, regions, assignment);
    hbboxWRSMTTM.stop();

    Timer hbbox0TM;
    HBBox0PlEval       hbbox0Eval  (_hgraphOfNetlist, regions, assignment);
    hbbox0TM.stop();
    Timer hbbox0WChengTM;
    HBBox0wChengPlEval hbbox0WCEval(_hgraphOfNetlist, regions, assignment);
    hbbox0WChengTM.stop();
    Timer hbbox0WRSMTTM;
    HBBox0wRSMTPlEval  hbbox0WREval(_hgraphOfNetlist, regions, assignment);
    hbbox0WRSMTTM.stop();

    Timer hbboxNTM;
    HBBoxNPlEval       hbboxNEval(_hgraphOfNetlist, regions, 
					assignment,_placement);
    hbboxNTM.stop();

    Timer hbboxNWChengTM;
    HBBoxNwChengPlEval  hbboxNWCEval(_hgraphOfNetlist, regions,
                                    	assignment, _placement);
    hbboxNWChengTM.stop();
    Timer hbboxNWRSMTTM;
    HBBoxNwRSMTPlEval  hbboxNWREval(_hgraphOfNetlist, regions,
                                    	assignment, _placement);
    hbboxNWRSMTTM.stop();

*/
    double bboxWL     = bboxEval.  getCost();
/*
    double bboxWCWL   = bboxWCEval.getCost();
    double bboxWRWL   = bboxWREval.getCost();
    double hbboxWL    = hbboxEval. getCost();
    double hbboxWCWL  = hbboxWCEval.getCost();
    double hbboxWRWL  = hbboxWREval.getCost();

    double hbbox0WL    = hbbox0Eval.getCost();
    double hbbox0WCWL  = hbbox0WCEval.getCost();
    double hbbox0WRWL  = hbbox0WREval.getCost();
    double hbboxNWL    = hbboxNEval.getCost();

    double hbboxNWCWL  = hbboxNWCEval.getCost();
    double hbboxNWRWL  = hbboxNWREval.getCost();

    if(_params.verb.forMajStats > 9)
    {
	cout<<"WL Estimation Results: (cost + runtimes)"<<endl;
    	cout<<"BBox         "<<bboxWL     <<bboxTM <<  endl;
        cout<<"BBoxWCheng   "<<bboxWCWL   <<bboxWChengTM << endl;
        cout<<"BBoxWRSMT    "<<bboxWRWL   <<bboxWRSMTTM << endl;

        cout<<"HBBox        "<<hbboxWL    <<hbboxTM << endl;
        cout<<"HBBoxWCheng  "<<hbboxWCWL  <<hbboxWChengTM << endl;
        cout<<"HBBoxWRSMT   "<<hbboxWRWL  <<hbboxWRSMTTM << endl;

        cout<<"HBBox0       "<<hbbox0WL   << hbbox0TM << endl;
        cout<<"HBBox0WCheng "<<hbbox0WCWL << hbbox0WChengTM << endl;
        cout<<"HBBox0WRSMT  "<<hbbox0WRWL << hbbox0WRSMTTM << endl;

        cout<<"HBBoxN       "<<hbboxNWL   << hbboxNTM << endl;
        cout<<"HBBoxNWCheng "<<hbboxNWCWL << hbboxNWChengTM << endl;
        cout<<"HBBoxNWRSMT  "<<hbboxNWRWL << hbboxNWRSMTTM << endl;
    }

    if(_params.verb.forMajStats > 2)
    {
	cout<<"WL Estimation Results: (cost + runtimes)"<<endl;
        cout<<"HBBoxN       "<<hbboxNWL   << hbboxNTM << endl;
    }
*/
    return bboxWL;
}


void CapoPlacer::replaceSmallBlocks()
{
    std::vector<CapoBlock*>::iterator blk;
    double prevWL, newWL;
 
    if(_params.verb.forMajStats > 0)
    {
      prevWL =  estimateWL();
      cout<<"\nEst WL before re-placing small blocks "<<prevWL<<endl;
    }

    vector<CapoBlock*> placedBlocksCpy = _placedBlocks;
    _placedBlocks.clear();

    for(blk = placedBlocksCpy.begin(); blk != placedBlocksCpy.end(); blk++)
      	refineBlock(**blk, *_curLayout, _layerNum,HAndV,0);
		//curLayout is a dummy var here..nothing will be
		//added to it, as these are all small blocks

    if(_params.verb.forMajStats > 0) 
    {
      newWL = estimateWL(); 
      cout<<"Est WL after re-placing small blocks "<<newWL<<endl;
    }

    for(blk= _placedBlocks.begin();blk != _placedBlocks.end(); blk++)
      delete *blk;
}


void CapoPlacer::setupAndCheck()
{
    vector<unsigned> coreCells;
    coreCells.reserve(_hgraphOfNetlist.getNumNodes());

   
    Orientation nOrient; //the 'default' N orientation
    itHGFNodeGlobal cell; 
    for(cell = _hgraphOfNetlist.nodesBegin();
	cell != _hgraphOfNetlist.nodesEnd(); cell++)
    {
        unsigned cellId = (*cell)->getIndex();

	if(_rbplace.isCoreCell(cellId))
	{ 
	    if(_rbplace.isFixed(cellId))  continue;

            coreCells.push_back(cellId);
	    
	    _placement.setOrient(cellId,nOrient);
	    abkassert((*cell)->getWeight() != 0, "core cell has 0 area ");
        }
	else  //must be fixed
	{
	    if(!_hgraphOfNetlist.isTerminal(cellId) && 
	        _params.verb.forMajStats > 3)
	    {
		cout<<"Non-Terminal is not a movable core cell"<<endl;
	 	cout<<"Cell "<<cellId<<" located at "<<_placement[cellId]<<endl;
	    }
	    abkassert(_rbplace.isFixed(cellId), "unfixed non-core cell");
	}
    }

    vector<const RBPCoreRow*> rows;
    vector<CROffset> endOffsets;
    itRBPCoreRow r;
    for(r = _rbplace.coreRowsBegin(); r != _rbplace.coreRowsEnd(); r++)
    {
	rows.push_back(r);
	const RBPSubRow& lastSR = *(r->subRowsBegin() + (r->getNumSubRows()-1));
	CROffset tmpOffset(r->getNumSubRows()-1, lastSR.getNumSites()-1);
	endOffsets.push_back(tmpOffset);
	_coreBBox += Point((*r)[0].getXStart(),r->getYCoord());
	_coreBBox += Point((*r)[endOffsets.back().first].getXEnd(),
			   r->getYCoord()+r->getHeight());
    }

    vector<CROffset> startOffsets(rows.size(), CROffset(0,0));

    if(_params.verb.forMajStats > 1)
	cout<<endl<<" Core BBox: "<<_coreBBox<<endl;

    vector<CapoBlock*> noNeighbors; 
    _layout[0].push_back(new 
                            CapoBlock(coreCells, _hgraphOfNetlist,
	   		    rows.begin(), rows.end(), startOffsets, endOffsets,
                            noNeighbors, noNeighbors, noNeighbors, noNeighbors,
			    (char*)"BLOCK"));

    CapoBlock* newBlk = _layout[0].back();
    Point center = newBlk->getCenter();

    for(unsigned c = 0; c < coreCells.size(); c++)
    {
	_cellToBlockMap[coreCells[c]] = newBlk;
	_placement[coreCells[c]] = center;
    }
                          
    _curLayout = &(_layout[0]);

    if(_params.verb.forMajStats > 4) printNetlistStats();
}

void CapoPlacer::updateInfoAboutBlock(const CapoBlock* block)
{
    std::vector<unsigned>::const_iterator cIt;
    Point blockCenter = block->getCenter();
    for(cIt  = block->cellIdsBegin(); cIt != block->cellIdsEnd(); cIt++)
    {
	_cellToBlockMap[*cIt] = block;
	_placement[*cIt] = blockCenter;
    }
}

void CapoPlacer::plotBlocks(unsigned layerNum, vector<CapoBlock*>& layout)
{
   double sum=0.0;
   unsigned b;
   for(b=0; b!=layout.size(); b++) sum+=layout[b]->getNumCells();
   if (sum / layout.size() < 10 ) return;

   if (layerNum>15) return;
   char fileName[100], baseName[100];
   sprintf(baseName,"capoBlocks-layer%d",layerNum);
   sprintf(fileName,"%s.gp",baseName);
   ofstream out(fileName);
   out << "# "<<layout.size() << " Capo blocks from layer " << layerNum << endl
       << "# Need gnuplot 3.7 or later version \n"
       << "#set noxtics \n#set noytics \nset noborder \nset nokey\n"
       << "set title 'Capo Layer " <<layerNum<<"'\n";
   for(b=0; b!=layout.size(); b++)
   {
     const CapoBlock& curBlk = *layout[b];
     Point ctr=curBlk.getBBox().getGeomCenter();
     out <<"set label '"<<int(curBlk.getRelativeWhitespace()*100)
// << "%/" << int(nodesToMove(curBlk)*100.0/curBlk.getCellIds().size())
   <<"%' at "<<ctr.x<<", "<<ctr.y<<"\n";
     Point weberLoc=getWeberLocation(curBlk);
     out <<"set arrow from "<<ctr.x<<", "<<ctr.y
                    <<" to "<<weberLoc.x<<","<<weberLoc.y<<endl;
   } 
   out << "# set terminal postscript\n# set output '"
       <<baseName<<".eps'"<<endl;
   out<< " plot '-' w l\n";
   for(b=0; b!=layout.size(); b++) 
   {
     xyDrawRectangle(out,layout[b]->getBBox());
     out << endl;
   }
   out << "EOF\n";
   out << "pause -1 'Press any key' "<<endl;
   
   cout << " Saved " << fileName << endl;
}

Point CapoPlacer::getPinLocation(unsigned cellId, unsigned netId) const
{
   if(!_params.useActualPinLocs || !_rbplace.isFixed(cellId))
       return _placement[cellId];
   else
   {
	return _placement[cellId] + 
	       _pinLocCalc.getPinOffset(cellId, netId);
   }
}


unsigned CapoPlacer::getTotalNetCut() const
{
    vector<int> nodeBlock;
    getBlockMembership(nodeBlock);
    unsigned numTotalCut = 0;
    itHGFEdgeGlobal edge;
    for(edge = _hgraphOfNetlist.edgesBegin();
        edge != _hgraphOfNetlist.edgesEnd(); edge++)
    {
        int blockID = -2;
        itHGFNodeLocal node;
        for(node = (*edge)->nodesBegin(); node != (*edge)->nodesEnd(); node++)
        {
            unsigned nodeId = (*node)->getIndex();

            // uncomment the following line if we don't want to count
            // nets that are connected to pads
            //if(nodeBlock[nodeId] == -1) continue;

            if(blockID == -2)
            {
                blockID = nodeBlock[nodeId];
                continue;
            }
            if (blockID != nodeBlock[nodeId])
            {
                numTotalCut ++;
                break;
            }
        }
    }
    return numTotalCut;
}

void CapoPlacer::getNetCutInfo(vector<unsigned>& externalCutNets,
                               vector<unsigned>& containedNets) const
{
    vector<int> nodeBlock;
    getBlockMembership(nodeBlock);
    unsigned numBlocks = _placedBlocks.size() + _curLayout->size();
    externalCutNets.clear();
    externalCutNets.reserve(numBlocks);
    containedNets.clear();
    containedNets.reserve(numBlocks);
    for(unsigned i = 0; i < numBlocks; i++)
    {
        externalCutNets.push_back(0);
        containedNets.push_back(0);
    }

    itHGFEdgeGlobal edge;
    BitBoard blockUsed(numBlocks);
    for(edge = _hgraphOfNetlist.edgesBegin();
        edge != _hgraphOfNetlist.edgesEnd(); edge++)
    {
        int blockID = -2;
        itHGFNodeLocal node;
        blockUsed.clear();
        for(node = (*edge)->nodesBegin(); node != (*edge)->nodesEnd(); node++)
        {
            unsigned nodeId = (*node)->getIndex();

            // uncomment the following line if we don't want to count
            // nets that are connected to pads
            //if(nodeBlock[nodeId] == -1) continue;

            if(blockID == -2)
            {
                blockID = nodeBlock[nodeId];
                continue;
            }
            if (blockID != nodeBlock[nodeId])
            {
                if(blockID >= 0 && !blockUsed.isBitSet(blockID))
                {
                    blockUsed.setBit(blockID);
                    externalCutNets[blockID]++;
                }
                if(nodeBlock[nodeId] >= 0 &&
                   !blockUsed.isBitSet(nodeBlock[nodeId]))
                {
                    blockUsed.setBit(nodeBlock[nodeId]);
                    externalCutNets[nodeBlock[nodeId]]++;
                }
            }
        }
        if(blockUsed.getNumBitsSet() == 1)
            containedNets[*(blockUsed.getIndicesOfSetBits().begin())] ++;
    }
}

void CapoPlacer::getBlockMembership(vector<int>& nodeBlock) const
{
    nodeBlock.clear();
    nodeBlock.reserve(_hgraphOfNetlist.getNumNodes());
    for(unsigned i = 0; i < _hgraphOfNetlist.getNumNodes(); i++)
        nodeBlock.push_back(-1);
    // Initialize the block number to -1 so that cells that are not
    // in any block will have -1 as its block id

    // cells that are in placed blocks
    int b, vectorSize = static_cast<int>(_placedBlocks.size());
    for(b = 0; b < vectorSize; b++)
    {
        const CapoBlock& curBlk = *(_placedBlocks[b]);
        std::vector<unsigned>::const_iterator nItr;
        for(nItr = curBlk.cellIdsBegin(); nItr != curBlk.cellIdsEnd(); nItr++)
            nodeBlock[*nItr] = b;
    }

    // cells that are in unplaced blocks
    // The block id continues from the placed blocks
    int baseID = static_cast<int>(_placedBlocks.size());
    vectorSize = static_cast<int>(_curLayout->size());
    for(b = 0; b < vectorSize; b++)
    {
        const CapoBlock& curBlk = *(*_curLayout)[b];
        std::vector<unsigned>::const_iterator nItr;
        for(nItr = curBlk.cellIdsBegin(); nItr != curBlk.cellIdsEnd(); nItr++)
            nodeBlock[*nItr] = b + baseID;
    }
}

void CapoPlacer::getHierCellNames(vector<const char*>& cellNames) const
{
    unsigned k;
    for(k=0; k!=cellNames.size(); k++)
     if (cellNames[k]) delete (char*)cellNames[k];
    cellNames.clear();
    cellNames.insert(cellNames.end(),_hgraphOfNetlist.getNumNodes(),
                     (char*)NULL);

    // cells that are in placed blocks
    unsigned vectorSize = _placedBlocks.size();
    bit_vector		seenCell(_hgraphOfNetlist.getNumNodes(), false);

    for(unsigned b = 0; b < vectorSize; b++)
    {
        const CapoBlock& curBlk = *(_placedBlocks[b]);
        const char* blkName=curBlk.getName();
        abkfatal(blkName," NULL block name in CapoPlacer::getHierCellName()"
        " --- make sure that small blocks were not replaced earlier\n");
        unsigned len=strlen(blkName)+20;
	char  *shortBlkName = new char[len];
	std::fill(shortBlkName, shortBlkName+len, 0);
	strncpy(shortBlkName, blkName, 17);	//only the first 4 levels
	//strcpy(shortBlkName, blkName);
	
cout<<"Shortened Block name: "<<shortBlkName<<endl;

        std::vector<unsigned>::const_iterator nItr;
        for(nItr = curBlk.cellIdsBegin(); nItr != curBlk.cellIdsEnd(); nItr++)
        {
	   abkfatal(!seenCell[*nItr], "already found this cell");
	   seenCell[*nItr] = true;
           cellNames[*nItr] = new char[len+10];
           sprintf((char*)cellNames[*nItr],"%s%d",shortBlkName,*nItr);
        }
    }

    // cells that are in unplaced blocks
    // The block id continues from the placed blocks
    for(k=0; k!=cellNames.size(); k++)
    {
        if (cellNames[k]) continue;
        const char* name=_hgraphOfNetlist.getNodeByIdx(k).getName();
        cellNames[k] = new char[strlen(name)+2];
        strcpy((char*)cellNames[k],name);
    }
}

void CapoPlacer::printBlockMembership(ostream& os) const
{
    os << "# every line below is of the form  \"<CellName> <BlockNumber>\"\n"
       << "# cells that do not belong to any block will have "
       << "BlockNumber = -1\n";
    if(_placedBlocks.size() > 0)
    os << "# blocks with BlockNumber from 0 to " << _placedBlocks.size() - 1
       << " are placed blocks\n";

    vector<int> nodeBlock;
    getBlockMembership(nodeBlock);

    itHGFNodeGlobal node;
    for(node = _hgraphOfNetlist.nodesBegin();
        node != _hgraphOfNetlist.nodesEnd(); node++)
            os << setw(50) << (*node)->getName() << "  " << setw(4)
               << nodeBlock[(*node)->getIndex()] << endl;
}             

void CapoPlacer::printBlockMembership(const char* fileName) const
{
    ofstream outFile(fileName);
    printBlockMembership(outFile);
    outFile.close();
}          

void CapoPlacer::printHierCellNames(ostream& os) const
{
    os << "# every line below is of the form  \"<CellName> <NewCellName>\"\n"
       << "# cells that do not belong to any block keep original names \n";
//     << "# cells that do not belong to any block will have "
//     << " NoBlock|<id> as name\n";

    vector<const char*> cellNames;
    getHierCellNames(cellNames);

    itHGFNodeGlobal node;
    for(node = _hgraphOfNetlist.nodesBegin();
        node != _hgraphOfNetlist.nodesEnd(); node++)
            os << setw(45) << (*node)->getName() << "  " << setw(25)
               << cellNames[(*node)->getIndex()] << endl; 
    for(unsigned k=0; k!=cellNames.size(); k++) delete (char*)cellNames[k];
}

void CapoPlacer::printHierCellNames(const char* fileName) const
{
    ofstream outFile(fileName);
    printHierCellNames(outFile);
}

