/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




//created by Andrew Caldwell 05/26/98

// CHANGES

// 980917 ilm   templated dcGainCo by GainBucketArray type
// 980918 ilm   added updateGainOffset methods

#ifndef __DEST_CENTRIC_GC_H__
#define __DEST_CENTRIC_GC_H__

#include "GainContainers/gainContainerXFace.h"
#include "GainContainers/regGainBucket.h"
#include "GainContainers/splitGainBucket.h"
#include "GainContainers/nByKRepository.h"
#include "svGainElmt.h"

#define DestCentricGainContainer DestCentricGainContainerT<BucketArray>
#define SplitDestCentricGainContainer \
	DestCentricGainContainerT<SplitBucketArray>

template <class BucketArrayT>
class DestCentricGainContainerT : public GainContainerInterface<SVGainElement>
{
  protected:

    NbyKRepository<SVGainElement>	_repository;
    vector<BucketArrayT>		_prioritizer;
	//there will be k BucketArrays (this is a k-way destination-
	//centric GainCont)

    unsigned	_highestArray;	//the array of buckets currently having
				//the highest valued gain in it.

    unsigned	_partitionBias;	//the gainCo will tie-break towards
				//moves which are TO this partition

    int			_maxGain;	//max possible gain a move can have
    const unsigned 	_numParts;


  public:

    DestCentricGainContainerT<BucketArrayT>(unsigned numNodes, 
					    unsigned numParts)
	: GainContainerInterface<SVGainElement>(),
	  _repository(numNodes, numParts),
	  _prioritizer(numParts), _highestArray(0),
	  _partitionBias(0), _maxGain(-1), _numParts(numParts)
	{}
  
    void setMaxGain(unsigned maxG);
 
    virtual void reinitialize();
            void setPartitionBias(unsigned part)
	 	{ _partitionBias = part;}

    void updateGain(SVGainElement& elmt, int deltaGain);
    void updateGain(unsigned node, unsigned part, int deltaGain);
    void updateGain(SVGainElement& elmt, unsigned part, int deltaGain);

    void updateGainOffset(SVGainElement& elmt, int deltaGain);
    void updateGainOffset(unsigned node, unsigned part, int deltaGain);
    void updateGainOffset(SVGainElement& elmt, unsigned part, int deltaGain);

/*  void  setGain(SVGainElement& elmt, int deltaGain);
    void  setGain(unsigned node, unsigned part, int deltaGain);
    void  setGain(SVGainElement& elmt, unsigned part, int deltaGain);

    void  setGainOffset(SVGainElement& elmt, int deltaGain);
    void  setGainOffset(unsigned node, unsigned part, int deltaGain);
    void  setGainOffset(SVGainElement& elmt, unsigned part, int deltaGain);
*/
    void  moveGainToOffset(SVGainElement& elmt);
    void  moveGainToOffset(unsigned node, unsigned part);
    void  moveGainToOffset(SVGainElement& elmt, unsigned part);

    void removeElement(SVGainElement& elmt) {    elmt.disconnect(); }
    void removeElement(unsigned nodeId, unsigned part)
	    {    _repository.getElement(nodeId, part).disconnect();}

    void removeAllNodesElements(unsigned nodeId)
    { for(unsigned k = 0; k < _numParts; k++)
	removeElement(getElement(nodeId, k));
    }

    int      getMaxGain()         const { return _maxGain;}
    unsigned getNumPrioritizers() const { return _prioritizer.size();}
    const BucketArrayT& getPrioritizer(unsigned k) const
    {	return _prioritizer[k];}
    const NbyKRepository<SVGainElement>& getRepository() const
    {   return _repository;}

    void removeAll();	//remove all the elements

    void addElement(SVGainElement& elmt, int gain);
    void addElement(SVGainElement& elmt, unsigned part, int gain);
    void addElement(unsigned nodeId, unsigned part, int gain)
	    {addElement(_repository.getElement(nodeId, part), part, gain);}

    SVGainElement& getElement(unsigned nodeId, unsigned part)
	    {    return _repository.getElement(nodeId, part);}
    unsigned getElementId(SVGainElement& elmt)
	    {	return _repository.getElementId(elmt);}
    unsigned getElementPartition(SVGainElement& elmt)
	    {	return _repository.getElementPartition(elmt);}
	
	    
	//I know these static casts are a 'Bad Thing' (tm), however,
	//the head of the lists MUST be derived from the base Element,
	//to allow for efficiently removing the item in lists of length
	//1.  It can't be templated, as you can't derive from a template 
	//parameter
 
    virtual SVGainElement* getHighestGainElement();

    virtual bool invalidateElement(SVGainElement& elmt);
    virtual bool invalidateBucket(SVGainElement& elmt);
					
	//these are somewhat faster, if the partition is already known..
    	    bool invalidateElement(unsigned part);
     	    bool invalidateBucket (unsigned part);

    virtual void resetAfterGainUpdate()
    {
        for(unsigned k = 0; k < _prioritizer.size(); k++)
	  _prioritizer[k].resetAfterGainUpdate();
        setHighestGainArray();
    }

	    void setupClip();

	    bool isConsistent();

  private:

    void setHighestGainArray()
    {		
        _highestArray  = _partitionBias;
    	int highestGain= _prioritizer[_highestArray].getHighestValidGain(); 
    	for(unsigned k = 0; k < _prioritizer.size(); k++)
    	 if(_prioritizer[k].getHighestValidGain() > highestGain)
    	 {
   	    _highestArray = k;
      	    highestGain = _prioritizer[k].getHighestValidGain();
    	 }
    }

};

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::updateGain
		(unsigned node, unsigned part, int deltaGain)
{ updateGain(_repository.getElement(node, part), part, deltaGain); }

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::
		updateGain (SVGainElement& elmt, int deltaGain)
{
    unsigned part = _repository.getElementPartition(elmt);
    updateGain(elmt, part, deltaGain);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::updateGain
		(SVGainElement& elmt, unsigned part, int deltaGain)
{
    elmt.disconnect();
    elmt._gain += deltaGain;
    _prioritizer[part].push(elmt);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::updateGainOffset
		(unsigned node, unsigned part, int deltaGain)
{
    updateGainOffset(_repository.getElement(node, part), part, deltaGain);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::
		updateGainOffset (SVGainElement& elmt, int deltaGain)
{
    unsigned part = _repository.getElementPartition(elmt);
    updateGainOffset(elmt, part, deltaGain);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::updateGainOffset
		(SVGainElement& elmt, unsigned part, int deltaGain)
{
    elmt.disconnect();
    elmt._gainOffset += deltaGain;
    _prioritizer[part].push(elmt);
}

/*
template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::setGain
		(unsigned node, unsigned part, int deltaGain)
{ setGain(_repository.getElement(node, part), part, deltaGain); }

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::
		setGain (SVGainElement& elmt, int deltaGain)
{
    unsigned part = _repository.getElementPartition(elmt);
    setGain(elmt, part, deltaGain);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::setGain
		(SVGainElement& elmt, unsigned part, int deltaGain)
{
    elmt.disconnect();
    elmt._gain += deltaGain;
    _prioritizer[part].push(elmt);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::setGainOffset
		(unsigned node, unsigned part, int deltaGain)
{
    setGainOffset(_repository.getElement(node, part), part, deltaGain);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::
		setGainOffset (SVGainElement& elmt, int deltaGain)
{
    unsigned part = _repository.getElementPartition(elmt);
    setGainOffset(elmt, part, deltaGain);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::setGainOffset
		(SVGainElement& elmt, unsigned part, int deltaGain)
{
    elmt.disconnect();
    elmt._gainOffset += deltaGain;
    _prioritizer[part].push(elmt);
}
*/

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::
     moveGainToOffset(unsigned node, unsigned part)
{
    moveGainToOffset(_repository.getElement(node, part), part);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::
     moveGainToOffset (SVGainElement& elmt)
{
    unsigned part = _repository.getElementPartition(elmt);
    moveGainToOffset(elmt, part);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::
     moveGainToOffset(SVGainElement& elmt, unsigned part)
{
    elmt.disconnect();
    elmt._gainOffset += elmt._gain;
    elmt._gain=0;
    _prioritizer[part].push(elmt);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::addElement
		(SVGainElement& elmt, int gain)
{    
    #ifdef ABKDEBUG
	if(elmt._next != NULL || elmt._prev != NULL)
	{
	  cout<<"trying to add a node that is already in the lists: ";
	  cout<<_repository.getElementId(elmt)<<endl;
	  abkassert(0, "done");
	}
    #endif

    elmt._gain = gain;
    _prioritizer[_repository.getElementPartition(elmt)].push(elmt);
}

template <class BucketArrayT>
inline void DestCentricGainContainerT<BucketArrayT>::addElement
		(SVGainElement& elmt, unsigned part, int gain)
{    
    #ifdef ABKDEBUG
	if(elmt._next != NULL || elmt._prev != NULL)
	{
	  cout<<"trying to add a node that is already in the lists: ";
	  cout<<_repository.getElementId(elmt)<<endl;
	  abkassert(0, "done");
	}
    #endif

    abkassert(part < _prioritizer.size(), "invalid partition in addGE");

    elmt._gain = gain;
    _prioritizer[part].push(elmt);
}

template <class BucketArrayT>
inline SVGainElement* DestCentricGainContainerT<BucketArrayT>::getHighestGainElement()
{
    return static_cast<SVGainElement*>
			(_prioritizer[_highestArray].getHighest()); 
}


template <class BucketArrayT>
inline bool DestCentricGainContainerT<BucketArrayT>::invalidateElement(SVGainElement& elmt)
{ 
    bool maxChanged;
    maxChanged = _prioritizer[_repository.getElementPartition(elmt)].
		invalidateElement();
    if(maxChanged)
	setHighestGainArray();
    return maxChanged;
}

template <class BucketArrayT>
inline bool DestCentricGainContainerT<BucketArrayT>::invalidateBucket(SVGainElement& elmt)
{  
    _prioritizer[_repository.getElementPartition(elmt)].invalidateBucket();
    setHighestGainArray();
    return true;
}

template <class BucketArrayT>
	//these are somewhat faster, if the partition is already known..
inline bool DestCentricGainContainerT<BucketArrayT>::invalidateElement(unsigned part)
{    
    bool maxChanged;
    maxChanged = _prioritizer[part].invalidateElement(); 
    if(maxChanged)
	setHighestGainArray();
    return maxChanged;
}

template <class BucketArrayT>
inline bool DestCentricGainContainerT<BucketArrayT>::invalidateBucket (unsigned part)
{ 
    _prioritizer[part].invalidateBucket(); 
    setHighestGainArray();
    return true;
}

template <class BucketArrayT>
ostream& operator<<
(ostream& os, const DestCentricGainContainerT<BucketArrayT>& gc)
{
    if(gc.getMaxGain() == -1)
	cout<<"setMaxGain has not been called. "
	    <<"Can't print DestCentricGainContainerT<BucketArrayT>";

    os<<"  MaxGain:  "<<gc.getMaxGain()<<endl;

    for(unsigned k = 0; k < gc.getNumPrioritizers(); k++)
    {
        os<<"Moves To Partition "<<k<<endl;

        for(int g = gc.getPrioritizer(k).size()-1; g >= 0; g--)
        {
            os << setw(4) << g - gc.getMaxGain() <<") ";
            SVGainElement* cur =
                static_cast<SVGainElement* const>(gc.getPrioritizer(k)[g]._next);
            if(cur == NULL)
                os<<" EMPTY "<<endl;
            else
            {
                for( ; cur != NULL;
                        cur = static_cast<SVGainElement* const>(cur->_next))
                {

                    os<<"[#:"<<gc.getRepository().getElementId(*cur);
                    os<<" g:"<<cur->getGain()<<" + ";
                    os<<cur->getRealGain() - cur->getGain()<<"]";

                }
                os<<endl;
            }
        }
    }
    return os;
}

#if defined (_MSC_VER) || (__GNUC__)
#define _INSIDE_DCGAINCONT_H
#include "dcGainCont.cxx"
#undef _INSIDE_DCGAINCONT_H
#endif

#endif

