/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




//created by Igor Markov on 06/02/98
#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#ifndef _FMMM_NC2w_CXX_
#define _FMMM_NC2w_CXX_

#include "fmMoveManagerNC2w.h"

void FMMoveManagerNC2w::computeGains()
{
  unsigned moduleIdx;

  if (!randomized)
  {
    for(moduleIdx=0; moduleIdx!=_numTerminals+_numMovable; moduleIdx++)
     computeGainsOfModule(moduleIdx); 
  }
  else
  { 
    Permutation randPerm(_numTerminals+_numMovable, Mapping::_Random);
    for(moduleIdx=0; moduleIdx!=_numTerminals+_numMovable; moduleIdx++)
     computeGainsOfModule(randPerm[moduleIdx]);
  }

if (wiggleNow)
{
  for(moduleIdx=0; moduleIdx!=_numTerminals+_numMovable; moduleIdx++)
  {
    if (_lockedModules[moduleIdx])
    {
        // touch gains of all neighbors
        const HGFNode& n = _problem.getHGraph().getNodeByIdx(moduleIdx);      
        for(itHGFEdgeLocal  e = n.edgesBegin(); e != n.edgesEnd(); e++)
        {
            const HGFEdge& net=(**e);
            for(itHGFNodeLocal gn=net.nodesBegin(); gn!=net.nodesEnd(); gn++)
            {
               unsigned gainingModuleIdx= (*gn)->getIndex();
               if (_lockedModules[gainingModuleIdx]) continue;
               if (gainingModuleIdx==moduleIdx) continue;
               unsigned to=1-(*_curPart)[gainingModuleIdx].lowestNumPart();
               _gainCo.updateGain(gainingModuleIdx,to,0);
            }
        }
    }
  }
  wiggleNow=false;
} 
  _gainCo.resetAfterGainUpdate();
  abkassert(_gainCo.isConsistent(), "GainCo inconsistent in computeGains");
}

void FMMoveManagerNC2w::uncutNets()
{
  cerr << " Running net uncutting " << endl;
  reinitialize();
  computeGains();  
// For all nets: uncut net if possible 
//cout << " Total cost before net uncutting " << _eval.getTotalCost() << endl;
  double maxLegalMoveAreaTo[2]={_partLeg.getMaxLegalMoveArea(1,0),
                                _partLeg.getMaxLegalMoveArea(0,1)};
// cout << " Is legal before uncutting " << _partLeg.isPartLegal() << endl;
//cout<<"MaxLegalMoveAreas:"<<maxLegalMoveAreaTo[0]<<" "<<maxLegalMoveAreaTo[1];
//cout<< endl;
  const HGraphFixed& hgraph = _problem.getHGraph();
  for(itHGFEdgeGlobal ee = hgraph.edgesBegin(); ee != hgraph.edgesEnd(); ee++)
  {            
     const HGFEdge& e=(**ee);
     if (e.getDegree()<5)  continue;
     if (e.getDegree()>30) continue;
     unsigned netIdx=e.getIndex();
     if (_eval.getNetCost(netIdx)==0) continue;
     bool tryMovingTo[2]={_eval.getTallies()[2*netIdx+1]>2,
                          _eval.getTallies()[2*netIdx]>2};
     if (!tryMovingTo[0] && !tryMovingTo[1]) continue;
     int gainSum[2]={0,0};
     maxLegalMoveAreaTo[0]=_partLeg.getMaxLegalMoveArea(1,0);
     maxLegalMoveAreaTo[1]=_partLeg.getMaxLegalMoveArea(0,1);
     double moveAreaSoFarTo[2]={0,0};

     itHGFNodeLocal  node;
     for(node = e.nodesBegin(); node != e.nodesEnd(); node++)
     {
        _movingModuleIdx= (*node)->getIndex();
        PartitionIds& curr   =(*_curPart)[_movingModuleIdx];
        _moveToPart = (curr.isInPart(0) ? 1 : 0);
        if (!tryMovingTo[_moveToPart]) continue;

        _moveFromPart = 1-_moveToPart;
        PartitionIds  canGoTo=(*_fixedConstr)[_movingModuleIdx];
        moveAreaSoFarTo[_moveToPart]+=(*node)->getWeight();
        if (!canGoTo[_moveToPart] ||  
             moveAreaSoFarTo[_moveToPart]>=maxLegalMoveAreaTo[_moveToPart])
        {
           tryMovingTo[_moveToPart]=false;
           if (tryMovingTo[_moveFromPart]) continue; else break; 
        }
        gainSum[_moveToPart]
               +=_gainCo.getElement(_movingModuleIdx,_moveToPart).getRealGain();
     }
     if (!tryMovingTo[0] && !tryMovingTo[1]) continue;
     unsigned moveNowTo=UINT_MAX;
     if      (!tryMovingTo[0])
     { 
        if (gainSum[1]<0) continue;
        moveNowTo=1; 
     }
     else if (!tryMovingTo[1])
     {
        if (gainSum[0]<0) continue;
        moveNowTo=0;
     }
     else
     {
          if (gainSum[0] > gainSum[1])
          {
              if (gainSum[0]<0) continue;
              moveNowTo=0;
          }
          else 
          {
             if (gainSum[1]<0)  continue;
             if (gainSum[1]>gainSum[0]) moveNowTo=1;          
             else moveNowTo=( _eval.getTallies()[2*netIdx]>
                         _eval.getTallies()[2*netIdx+1] ? 1 : 0 );
          }
    }

     // now UNCUT the net by moving all nodes into _moveToPart
     cout << "Uncutting net " << netIdx << " of degree "
          << e.getDegree() << " into " << moveNowTo
          << " with sum of gains " << gainSum[moveNowTo] ;
     unsigned willMove=_eval.getTallies()[2*netIdx+(1-moveNowTo)]; 
     cout << "; will move " << willMove << " nodes "<< endl;
/*
     cout << "Tallies: " << _eval.getTallies()[2*netIdx] << " " <<
                          _eval.getTallies()[2*netIdx+1];
     cout << " tryMovingTo: " << tryMovingTo[0]<<" "<< tryMovingTo[1] << endl;
*/
//   cout << " Max legal move areas : " <<
//           maxLegalMoveAreaTo[0]  << " " << maxLegalMoveAreaTo[1] ;
//   cout << " this move " << moveAreaSoFarTo[0]<<" "<<moveAreaSoFarTo[1]<<endl;
     unsigned nodesMoved=0;
     for(node = e.nodesBegin(); node != e.nodesEnd(); node++)
     {
        _movingModuleIdx= (*node)->getIndex();
        PartitionIds& curr   =(*_curPart)[_movingModuleIdx];
        _moveToPart = (curr.isInPart(0) ? 1 : 0);
        if ( _moveToPart != moveNowTo ) continue;
        if (nodesMoved++==willMove) break;
        _moveFromPart = 1-_moveToPart;
        updateGains();
        (*_curPart)[_movingModuleIdx].setToPart(_moveToPart);
        _eval.moveModuleTo   (_movingModuleIdx,_moveFromPart,_moveToPart);
        _partLeg.moveModuleTo(_movingModuleIdx,_moveFromPart,_moveToPart);
     }
//  abkfatal(_partLeg.isPartLegal(), " Illegal partition ");
//   cout << " actually moved " << nodesMoved << endl;
  }
// cout << " Total cost after net uncutting " << _eval.getTotalCost() << endl;
// cout << " Is legal after uncutting " << _partLeg.isPartLegal() << endl;
//cout<<"MaxLegalMoveAreas:"<<maxLegalMoveAreaTo[0]<<" "<<maxLegalMoveAreaTo[1];
// cout << "\n-----------" << endl;
  _gainCo.resetAfterGainUpdate();
}

#define FMMMclass  FMMoveManagerNC2w

#define updateGainOfModule( moduleIdx,  from, to,  deltaGain) \
	 _gainCo.updateGain(moduleIdx,to,deltaGain)

#include "_pick_NC2w.cxx"
#include "_updateA_NC2w.cxx"

#undef FMMMclass
#undef updateGainOfModule

#endif
