/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

// Created by Igor Markov 980320

// CHANGES
// 980325  ilm  independent fix and blk files, AUX file I/O, stream I/O
// 980402  ilm  .are files are optional now

#include <fstream.h>
#include <bvector.h>
#include "partProb.h"
#include "partitionData.h"

PartitioningProblem::PartitioningProblem(const char* baseFileName,Verbosity v)
: _verb(v)
{
    initToNULL();
    read(baseFileName);
    _postProcess();
}

PartitioningProblem:: PartitioningProblem( const char* netDFileName,
                                           const char* areFileName, 
                                           const char* blkFileName,
                                           const char* fixFileName,
                                           const char* solFileName,
                                           Verbosity   verbosity )
       : _verb(verbosity)
{
   Timer tm;
   initToNULL();
   bool success=read(netDFileName,areFileName,
                             blkFileName,fixFileName,solFileName);
   abkwarn(success, "Reading PartitioningProblem from files may have failed");
   _postProcess();
   tm.stop();
   if (_verb.forSysRes) 
      cout << "Reading Partitioning Problem took " << tm << endl;
}


bool PartitioningProblem::read(const char * baseFileName)
{
    Timer tm;
    unsigned baseLength = strlen (baseFileName);
    ifstream auxFile(baseFileName);
    bool success=false;
    char * newNetD = NULL;
    char * newAre  = NULL;
    char * newBlk  = NULL;
    char * newFix  = NULL;
    char * newSol  = NULL;

    char dir[255];
    char *auxFileDirEnd=strrchr(baseFileName,pathDelim);
    if (auxFileDirEnd)
    {
       strncpy(dir,baseFileName,auxFileDirEnd-baseFileName+1);
       dir[auxFileDirEnd-baseFileName+1]=0;
    }
    else strcpy(dir,"");    

    if (auxFile)
    {
       bool found=false;
       int  lineNo=1;
       cout << "Reading " << baseFileName << endl;
       char oneLine[1023];
       char word1[100],word2[100];

       while (!found && !auxFile.eof())
       {
          auxFile >> eathash(lineNo) >> word1
                >> noeol(lineNo) >> word2 >> noeol(lineNo);
          abkfatal(!strcmp(word2,":"),
              " Error in aux file: space-separated column expected");
          if (!strcasecmp(word1,"CD"))
          {
              auxFile >> word1;
              auxFile >> needeol(lineNo++);
              if (word1[0]==pathDelimWindows || word1[0] == pathDelimUnix)
                  strcpy(dir,word1);
              else strcat(dir,word1);
              char     fDel[2];
              sprintf(fDel,"%c",pathDelim);
              if (word1[strlen(word1)-1]!=pathDelimWindows
                  || word1[0] == pathDelimUnix)
                  strcat(dir,fDel);
          }        
          else if (!strcasecmp(word1,"PartProb"))
          {
              found=true;       
              auxFile.getline(oneLine,1023);
              unsigned len=strlen(oneLine), fileNum=0;
              bool space=true;
              char *fileNames[5] = {  NULL, NULL,NULL, NULL, NULL };
              for(unsigned j=0; j!=len; j++)
              {
                 if (isspace(oneLine[j]))
                 {
                     if (!space) oneLine[j]='\0';
                     space=true;
                 }
                 else if (space)
                 {
                     space=false;
                     abkfatal(fileNum<5," Too many filenames in AUX file");
                     fileNames[fileNum++]=oneLine+j;
                 } 
              }

              for(j=0; j!=fileNum; j++)
              {
                 if (strstr(fileNames[j],".net")) newNetD=fileNames[j];   
                 if (strstr(fileNames[j],".NET")) newNetD=fileNames[j];   
                 if (strstr(fileNames[j],".are")) newAre=fileNames[j];   
                 if (strstr(fileNames[j],".ARE")) newAre=fileNames[j];   
                 if (strstr(fileNames[j],".blk")) newBlk=fileNames[j];   
                 if (strstr(fileNames[j],".BLK")) newBlk=fileNames[j];   
                 if (strstr(fileNames[j],".fix")) newFix=fileNames[j];   
                 if (strstr(fileNames[j],".FIX")) newFix=fileNames[j];   
                 if (strstr(fileNames[j],".sol")) newSol=fileNames[j];   
                 if (strstr(fileNames[j],".SOL")) newSol=fileNames[j];   
              }

              char * tmpNet=NULL;
              char * tmpAre=NULL;
              char * tmpBlk=NULL;
              char * tmpFix=NULL;
              char * tmpSol=NULL;
      
              unsigned dirLen=strlen(dir);

              if (newNetD!=NULL)
              if (newNetD[0]!=pathDelimWindows && newNetD[0] != pathDelimUnix )
              {
                  tmpNet=new char[dirLen+strlen(newNetD)+1];
                  sprintf(tmpNet,"%s%s",dir,newNetD);
              }
              else
              {
                  tmpNet=new char[strlen(newNetD)+1];
                  strcat(tmpNet,newNetD);
              }       

              if (newAre!=NULL)
              if  ( newAre[0]!=pathDelimWindows && newAre[0] != pathDelimUnix )
              {
                  tmpAre=new char[dirLen+strlen(newAre)+1];
                  sprintf(tmpAre,"%s%s",dir,newAre);
              }        
              else
              {
                  tmpAre=new char[strlen(newAre)+1];
                  strcat(tmpAre,newAre);
              }       
             
              if (newBlk!=NULL)
              if (newBlk[0]!=pathDelimWindows && newBlk[0] != pathDelimUnix )
              {
                  tmpBlk=new char[dirLen+strlen(newBlk)+1];
                  sprintf(tmpBlk,"%s%s",dir,newBlk);
              }        
              else
              {
                  tmpBlk=new char[strlen(newBlk)+1];
                  strcat(tmpBlk,newBlk);
              }       
             
              if (newFix!=NULL)
              if (newFix[0]!=pathDelimWindows && newFix[0] != pathDelimUnix )
              {
                  tmpFix=new char[dirLen+strlen(newFix)+1];
                  sprintf(tmpFix,"%s%s",dir,newFix);
              }        
              else
              {
                  tmpFix=new char[strlen(newFix)+1];
                  strcat(tmpFix,newFix);
              }       

              if (newSol!=NULL)
              if (newSol[0]!=pathDelimWindows && newSol[0] != pathDelimUnix )
              {
                  tmpSol=new char[dirLen+strlen(newSol)+1];
                  sprintf(tmpSol,"%s%s",dir,newSol);
              }        
              else
              {
                  tmpSol=new char[strlen(newSol)+1];
                  strcat(tmpSol,newSol);
              }       
               
               
              success=read(_abk_cpd(tmpNet),_abk_cpd(tmpAre),
                   _abk_cpd(tmpBlk),_abk_cpd(tmpFix), _abk_cpd(tmpSol));

              if (tmpNet) delete[] tmpNet;
              if (tmpAre) delete[] tmpAre;
              if (tmpBlk) delete[] tmpBlk;
              if (tmpFix) delete[] tmpFix;
              if (tmpSol) delete[] tmpSol;
           }     
        }
    }
    else
    {
         newNetD = new char[baseLength + 6];
         newAre  = new char[baseLength + 5];
         newBlk  = new char[baseLength + 5];          
         newFix  = new char[baseLength + 5];          
         newSol  = new char[baseLength + 5];          
         strcpy(newNetD, baseFileName); strcat(newNetD, ".netD");
         strcpy(newAre,  baseFileName); strcat(newAre,  ".are");
         strcpy(newBlk,  baseFileName); strcat(newBlk,  ".blk");
         strcpy(newFix,  baseFileName); strcat(newFix,  ".fix");              
         strcpy(newSol,  baseFileName); strcat(newSol,  ".sol");              
 
         abkfatal2(ifstream(newNetD)," No .aux, net or .netD found: ",
                                      baseFileName);
         success=read(newNetD,newAre, newBlk, newFix, newSol);
  
         delete[] newNetD;
         delete[] newAre;
         delete[] newBlk;
         delete[] newFix;
    }
    abkwarn(success,"Reading PartitioningProblem from files may have failed\n");
    tm.stop();
    if (_verb.forSysRes) 
       cout << "Reading PartitioningProblem took " << tm << endl;
    return true;
}

bool PartitioningProblem::save(const char* baseFileName)
{
    unsigned baseLength = strlen (baseFileName);
    char * newAux  = new char[baseLength + 6];
    char * newNetD = new char[baseLength + 6];
    char * newAre  = new char[baseLength + 5];
    char * newBlk  = new char[baseLength + 5];          
    char * newFix  = new char[baseLength + 5];          
    char * newSol  = new char[baseLength + 5];          
    strcpy(newAux,  baseFileName); strcat(newAux, ".aux");
    strcpy(newNetD, baseFileName); strcat(newNetD, ".netD");
    strcpy(newAre,  baseFileName); strcat(newAre,  ".are");
    strcpy(newBlk,  baseFileName); strcat(newBlk,  ".blk");
    strcpy(newFix,  baseFileName); strcat(newFix,  ".fix");              
    strcpy(newSol,  baseFileName); strcat(newSol,  ".sol");              

    if (_terminalToBlock==NULL) 
    {
       delete[] newFix;
       newFix=NULL;
    }
    if (_partitions==NULL)
    {
       delete[] newBlk;
       delete[] newSol;
       newBlk=NULL, newSol=NULL;
    }
    else if (_bestSolnNum>=_solnBuffers->size() && newSol) 
    { 
       delete[] newSol;
       newSol=NULL;
    }
 
    bool success=save(newNetD,newAre, newBlk, newFix, newSol);
    abkwarn(success, "Reading PartitioningProblem from files may have failed");
    if (success)
    {
       ofstream auxFile(newAux);
       abkfatal(newAux," Can't open AUX file for writing");
       auxFile << "PartProb : " << newNetD << " " << newAre << " ";
       if (newBlk) auxFile << newBlk << " ";
       if (newFix) auxFile << newFix << " ";
       if (newSol) auxFile << newSol << " ";
       auxFile << endl;
    }

    delete[] newAux;
    delete[] newNetD;
    delete[] newAre;
    if (newBlk) delete[] newBlk;
    if (newFix) delete[] newFix;
    if (newSol) delete[] newSol;
    return success;
}

bool PartitioningProblem 
::save(const char* netDFileName, const char* areFileName,
       const char* blkFileName, const char* fixFileName,
       const char* solFileName) const
{
   abkfatal(netDFileName,  
       " Can't save PartitioningProblem: NULL .netD file name passed");
   abkfatal(areFileName,  
       " Can't save PartitioningProblem: NULL .are file name passed");

   bool success=_hgraph->saveAsNetDAre(netDFileName,areFileName);
   abkwarn(success," Writing into .netD and .are files may have failed");
   cout << " Done writing " << netDFileName << " and " << areFileName << endl;

   abkfatal(_hgraph," No hypergraph found. Can't save partitioning problem.");
   abkfatal(_totalWeight," Partitioning problem corrupt. Can't save.");

   unsigned k,j,numWeights=(*_totalWeight).size();
   vector<double> sumCapacities(numWeights,0);

   if (_capacities==NULL) return true;

   unsigned numPart=(*_capacities).size();

   char buf[20],buf1[20];

   for(k=0; k!=numPart; k++)
   {    
      for(j=0;j!=numWeights;j++) sumCapacities[j]+=(*_capacities)[k][j];
   } 

   if ( blkFileName && _padBlocks && _partitions)
   {
      unsigned numPadBlocks=(*_padBlocks).size();

      cout << " Writing " << blkFileName << " ... " << endl;
      ofstream blkFile(blkFileName);
      abkfatal(blkFile," Can't open .blk file for writing");

                                                              
      blkFile << "UCLA blk 1.0 " << endl 
           << TimeStamp() << User() << Platform() << endl;

//  Write .blk file

   blkFile << "Regular partitions : " << numPart       << endl
           << "Pad partitions : "     << numPadBlocks  << endl
           << "Relative capacities : ";
   for (k=0; k!=(*_totalWeight).size(); k++) blkFile << "yes ";
   blkFile << endl << "Capacity tolerances : ";

   const vector<double>& firstNodeCaps=(*_capacities)[0];
   const vector<double>& firstNodeMaxCaps=(*_maxCapacities)[0];
   vector<double> capTols(firstNodeCaps.size());
   for (k=0; k!=capTols.size(); k++) 
   {
       capTols[k]=100.0*abs(firstNodeMaxCaps[k]- firstNodeCaps[k])
                   /firstNodeCaps[k];         // sumCapacities[k] ; 
       blkFile << capTols[k] << "%   ";
   }
   blkFile <<  endl;
   for(k=0;k!=_padBlocks->size();k++)
   {
      sprintf(buf ,"pb%d",k);
      sprintf(buf1,"%8s rect ",buf);
      blkFile << buf1;
      const Rectangle& rect=(*_padBlocks)[k];
      if (rect.isEmpty()) blkFile << "        Empty ";
      else 
      {
         blkFile << setw(10) << rect.xMin << " " 
                 << setw(10) << rect.yMin << " "
                 << setw(10) << rect.xMax << " " 
                 << setw(10) << rect.yMax << " ";
      }
//    blkFile << " : ";
      blkFile << endl;
   }

   for(k=0;k!=numPart;k++)
   {
      sprintf(buf ,"b%d",k);
      sprintf(buf1,"%8s rect ",buf);
      blkFile << buf1;
      const Rectangle& rect=(*_partitions)[k];
      if (rect.isEmpty()) blkFile << "        Empty ";
      else 
      {
         blkFile << setw(10) << rect.xMin << " " 
                 << setw(10) << rect.yMin << " "
                 << setw(10) << rect.xMax << " " 
                 << setw(10) << rect.yMax << " ";
      }
      blkFile << " : ";
      const vector<double>& caps=(*_capacities)[k];
      for(j=0; j!=caps.size(); j++)
      {
         blkFile << setw(10) << 100*caps[j]/sumCapacities[j] << " "; 
// compute cap tols and write them here if difft from capTols
      }
      blkFile << endl;
   }
   }

//  Now write .fix file

   if (fixFileName && _terminalToBlock)
   {
   ofstream fixFile(fixFileName); 
   abkfatal(fixFile," Can't open .fix file for writing");
   cout << " Writing " << fixFileName << " ... " << endl;
   fixFile << "UCLA fix 1.0 " << endl 
           << TimeStamp() << User() << Platform() << endl;

   PartitionIds freeToMove;
   freeToMove.setToAll(numPart);

   unsigned numFixedPads=0;
   unsigned numTermsFixedInBlocks= ( _terminalToBlock 
           ? min(_hgraph->getNumTerminals(), _terminalToBlock->size()) : 0 ); 
   for(k=0;k!=numTermsFixedInBlocks;k++)
   {
      if ((*_terminalToBlock)[k]!=UINT_MAX || (*_fixedConstr)[k]!=freeToMove)
      numFixedPads++;
   }

   unsigned numFixedNonPads=0;
   for(k=numTermsFixedInBlocks; k!=_hgraph->getNumNodes();k++)
   {
      PartitionIds constrainedTo=(*_fixedConstr)[k];
      if (constrainedTo!=freeToMove) numFixedNonPads++;
   }

   fixFile << "Regular partitions : " << (*_partitions).size() << endl
           << "Pad partitions : "     << (*_padBlocks).size()  << endl;
   fixFile << "Fixed Pads : " << numFixedPads << endl;
   fixFile << "Fixed NonPads : " << numFixedNonPads << endl;
 
   unsigned numtermsfixedinblocks= ( _terminaltoblock 
           ? min(_hgraph->getnumterminals(), _terminaltoblock->size()) : 0 ); 
   for(k=0;k!= numtermsfixedinblocks; k++)
   {
      unsigned pb=(*_terminaltoblock)[k];
      partitionids propagatedto=(*_fixedconstr)[k];
      if ( pb!=uint_max || propagatedto!=freetomove )
      {
         sprintf(buf ,"p%d",k+1);
         sprintf(buf1,"%8s : ",buf);
         fixfile << buf1;
         if ((*_terminaltoblock)[k]!=uint_max)  fixfile << "pb" << pb << " ";
         if ( propagatedto != freetomove )
         for (j=0; j!=8*sizeof(partitionids); j++)
           if (propagatedto[j])  fixfile << "b" << j << " ";
        fixfile << endl; 
      } 
   }

   unsigned modulenum=0;
   for(k= numtermsfixedinblocks; k!=_hgraph->getnumnodes();k++, modulenum++)
   {
      partitionids constrainedto=(*_fixedconstr)[k];
      if ( constrainedto!=freetomove)
      {
         sprintf(buf ,"a%d",modulenum);
         sprintf(buf1,"%8s : ",buf);
         fixfile << buf1;
         if (constrainedto!=freetomove)
         for (j=0; j!=8*sizeof(partitionids); j++)
           if (constrainedto[j])  fixfile << "b" << j << " ";
        fixfile << endl; 
      } 
   }
   }

   if (solfilename && _bestsolnnum < _solnbuffers->size())
       savebestsol(solfilename);
   return true;
}

bool partitioningproblem::savebestsol(const char* solfilename) const
{
   unsigned bestsol=getbestsolnnum();
   ofstream solfile(solfilename); 
   abkfatal(solfile," can't open .sol file for writing");
   cout << " writing " << solfilename << " ... " << endl;
   solfile << "ucla sol 1.0 " << endl 
           << timestamp() << user() << platform() << endl;

   unsigned numpart=(*_capacities).size();
   partitionids freetomove, nowhere;
   freetomove.settoall(numpart);

   unsigned j,k,numfixedpads=0;

   for(k=0;k!=_hgraph->getnumterminals();k++)
   {
      if ((*_solnBuffers)[bestSol][k]!=freeToMove)
      numFixedPads++;
   }
   unsigned numTermsFixedInBlocks= ( _terminalToBlock 
           ? min(_hgraph->getNumTerminals(), _terminalToBlock->size()) : 0 ); 

   unsigned numFixedNonPads=0;
   for(k=numTermsFixedInBlocks;k!=_hgraph->getNumNodes();k++)
   {
      PartitionIds constrainedTo=(*_solnBuffers)[bestSol][k];
      if (constrainedTo!=freeToMove) numFixedNonPads++;
   }

   solFile << "Regular partitions : " << (*_partitions).size() << endl
           << "Pad partitions : "     << (*_padBlocks).size()  << endl;
   solFile << "Fixed Pads : " << numFixedPads << endl;
   solFile << "Fixed NonPads : " << numFixedNonPads << endl;

   char buf[20],buf1[20];
  
   for(k=0;k!=numTermsFixedInBlocks;k++)
   {
      PartitionIds propagatedTo=(*_solnBuffers)[bestSol][k];
      if ( propagatedTo!=freeToMove )
      {
         sprintf(buf ,"p%d",k+1);
         sprintf(buf1,"%8s : ",buf);
         solFile << buf1;
         if ( propagatedTo != freeToMove )
         for ( j=0; j!=8*sizeof(PartitionIds); j++)
           if (propagatedTo[j])  solFile << "b" << j << " ";
        solFile << endl; 
      } 
   }

   unsigned moduleNum=0;
   for(k=numTermsFixedInBlocks;k!=_hgraph->getNumNodes();k++,moduleNum++)
   {
      PartitionIds constrainedTo=(*_solnBuffers)[bestSol][k];
      if ( constrainedTo!=freeToMove)
      {
         sprintf(buf ,"a%d",moduleNum);
         sprintf(buf1,"%8s : ",buf);
         solFile << buf1;
         if (constrainedTo!=freeToMove)
         for (j=0; j!=8*sizeof(PartitionIds); j++)
           if (constrainedTo[j])  solFile << "b" << j << " ";
        solFile << endl; 
      } 
   }
   return true;  
}

bool PartitioningProblem ::readBLK(istream& blkFile)
{
      abkfatal(_hgraph!=0," _hgraph is not yet initialized in PartProb");
      abkfatal(blkFile," Can't open .blk file for reading");

      unsigned k, j, numWeights=_hgraph->getNumWeights();
      unsigned numTerminals=_hgraph->getNumTerminals();
      unsigned numNodes=_hgraph->getNumNodes();

      vector<double> minNodeWeights(numWeights,DBL_MAX);
      for (itHGFNodeGlobal v=(*_hgraph).nodesBegin();
                        v!=(*_hgraph).nodesEnd(); ++v)  
      {
          for(k=0;k!=numWeights;k++)
          {
             double w=(*v)->getWeight(k);
             if (w<minNodeWeights[k]) minNodeWeights[k]=w;
          }
      }

      int lineNo=1;
      unsigned numPart=0, numPadBlocks=0;
      blkFile >> eathash(lineNo) >> needcaseword("UCLA",lineNo) 
              >> needcaseword("blk",lineNo) >> skiptoeol; 
      lineNo++;
      blkFile >> eathash(lineNo) >> needcaseword("Regular",lineNo)
              >> needcaseword("partitions", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numPart >> needeol(lineNo);

      _partitions=new vector<BBox>(numPart);
      PartitionIds freeToMove;
      freeToMove.setToAll(numPart);
      
      if (_fixedConstr==NULL)
      _fixedConstr=new Partitioning((*_hgraph).getNumNodes(),freeToMove);

      _capacities=new vector<vector<double> >
                                  (numPart,vector<double>(numWeights,-1));
      _maxCapacities=new vector<vector<double> >
                                  (numPart,vector<double>(numWeights,-1));
      _minCapacities=new vector<vector<double> >
                                  (numPart,vector<double>(numWeights,-1));
      blkFile >> eathash(lineNo) >> needcaseword("Pad",lineNo)
              >> needcaseword("partitions", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numPadBlocks >> needeol(lineNo);

      _padBlocks=new vector<BBox>(numPadBlocks);

      blkFile >> eathash(lineNo) >> needcaseword("Relative",lineNo)
              >> needcaseword("Capacities", lineNo) >> needword(":",lineNo);

      bit_vector relativeCaps(numWeights);
      for(k=0; k!=numWeights; k++)
      {
         char yesno[255];
         blkFile >> noeol(lineNo) >> isword(lineNo) >> yesno;
         if (strcasecmp(yesno,"yes")==0 || strcasecmp(yesno,"y")==0)
            relativeCaps[k]=true;
         else if (strcasecmp(yesno,"no")==0 || strcasecmp(yesno,"n")==0)
            relativeCaps[k]=false;
         else abkfatal2(0,"'yes' or 'no' expected in line ", lineNo);
      }
      blkFile >> needeol(lineNo);

      blkFile >> eathash(lineNo) >> needcaseword("Capacity",lineNo)
              >> needcaseword("tolerances", lineNo) >> needword(":",lineNo);

      vector<double> tolerances(numWeights);
      vector<unsigned> tolType(numWeights);
      for(k=0; k!=numWeights; k++)
      {
         char *qualif=NULL;
         char toler[255];
         blkFile >> noeol(lineNo) >> isnumber(lineNo) >> toler;
         if ( (qualif=strchr(toler,'b'))!=NULL )
         {
              qualif[0]='\0';
              // tolerance in terms of min node width
              tolType[k]=2;
              double fact=strtod(toler, &qualif);
              abkfatal2(fact>=1,
               " Capacity tolerance less than min node weight in line ",lineNo);
              tolerances[k]=fact*minNodeWeights[k];
              abkfatal2(qualif!=toler, " Expected a number in line ", lineNo);
         }
         else if ( (qualif=strchr(toler,'%'))!=NULL ) 
         {
              qualif[0]='\0';
              // tolerance in relative terms
              tolType[k]=0;
              tolerances[k]=0.01*strtod(toler, &qualif);
              abkfatal2(qualif!=toler, " Expected a number in line ", lineNo);
         }         
         else  
         {
              tolerances[k]=strtod(toler, &qualif);
              tolType[k]=1;
              abkfatal2(qualif!=toler, " Expected a number in line ", lineNo);
              abkfatal2((qualif-toler)==strlen(toler), 
                  " Trailing characters near tolerance in line ", lineNo);
         }
      }
      for(k=0; k!=numPart+numPadBlocks; k++)
      {
         char partId[256], rect[256];
         blkFile >> eathash(lineNo) >> isword(lineNo) 
                 >> partId >> noeol(lineNo) >> rect >> noeol(lineNo);

        abkfatal2(strcasecmp(rect,"rect")==0,"'rect' expected in line",lineNo);
         double xMin, yMin, xMax, yMax;
         blkFile >> isnumber(lineNo) >> xMin
                 >> isnumber(lineNo) >> yMin
                 >> isnumber(lineNo) >> xMax
                 >> isnumber(lineNo) >> yMax;
         BBox box(xMin,yMin,xMax,yMax);

         if (partId[0]=='b')
         {
             int partNum=atoi(partId+1);
             abkfatal2(partNum>=0 && partNum < numPart,
                 " Bad partition Id in line ", lineNo);
             (*_partitions)[partNum]=box;
             blkFile >> needword(":",lineNo);
             for(j=0;j!=numWeights;j++)
             {
                 blkFile >> noeol(lineNo) >> isnumber(lineNo) 
                         >> (*_capacities)[partNum][j];
             }
         }
         else if (partId[0]=='p' && partId[1]=='b')
         {
             int padBlockNum=atoi(partId+2);
             abkfatal2(padBlockNum>=0 && padBlockNum < numPadBlocks,
                 " Bad pad block Id (too big a number) in line ", lineNo);
             (*_padBlocks)[padBlockNum]=box;
         }
         else 
         {
            char buf[255];
            sprintf(buf,"\n Got '%s' ",partId);
            abkfatal3(0, "Partition Id starting with b or pb expected in line "
                       ,lineNo,buf);
         }
         blkFile >> needeol(lineNo);
      }
      for(k=0;k!=numWeights;k++)
      {
         if (relativeCaps[k]) 
         for (j=0;j!=numPart;j++)
         {
             (*_capacities)[j][k]*=0.01*(*_totalWeight)[k];
             switch (tolType[k])
             {
                case 0:
                 (*_maxCapacities)[j][k]=(*_capacities)[j][k]*(1+tolerances[k]);
                 (*_minCapacities)[j][k]=(*_capacities)[j][k]*(1-tolerances[k]);
                 break;
                case 1: case 2: 
                 (*_maxCapacities)[j][k]=(*_capacities)[j][k]+tolerances[k];
                 (*_minCapacities)[j][k]=(*_capacities)[j][k]-tolerances[k];
                 break;
                default: abkfatal(0,"Unknown balance tolerance type");
             }
         }
      }
  return true;
}

bool PartitioningProblem ::readFIX(istream& fixFile)
{
      abkfatal(_hgraph!=0," _hgraph is not yet initialized in PartProb");
      abkfatal(fixFile," Can't open .fix file for reading");

      unsigned k,numWeights=_hgraph->getNumWeights();
      unsigned numTerminals=_hgraph->getNumTerminals();
      unsigned numNodes=_hgraph->getNumNodes();

      abkfatal(_terminalToBlock==0, " _terminalToBlock is already init'd");
      _terminalToBlock=new vector<unsigned>(numTerminals,UINT_MAX);

      int lineNo=1;
      unsigned numPart=0, numPadBlocks=0, numFixedPads=0, numFixedNonPads=0;

      fixFile >> eathash(lineNo);
      fixFile >> needcaseword("UCLA",lineNo) 
              >> needcaseword("fix",lineNo) ;
      fixFile >> skiptoeol;
      lineNo++;

      fixFile >> eathash(lineNo);
      fixFile >> needcaseword("Regular",lineNo)
              >> needcaseword("partitions", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numPart >> needeol(lineNo);

      abkfatal(numPart>1,  " Need at least two regular partitions");
      abkfatal(numPart<=32," Can't have more than 32 regular partitions");

      if (_partitions!=NULL)
      {
          abkfatal3((*_partitions).size()==numPart,
              "Mismatch with #partitions from blk file: line ",lineNo,"\n");
      }

      if (_capacities!=NULL)
      {
          abkfatal2((*_capacities).size()==numPart,
              "Mismatch with #partitions from blk file: line ",lineNo);
      }
      else _capacities=new vector< vector<double> > (numPart);

      fixFile >> eathash(lineNo) >> needcaseword("Pad",lineNo)
              >> needcaseword("partitions", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numPadBlocks >> needeol(lineNo);

      fixFile >> eathash(lineNo) >> needcaseword("Fixed",lineNo)
              >> needcaseword("Pads", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numFixedPads >> needeol(lineNo);
      fixFile >> eathash(lineNo) >> needcaseword("Fixed",lineNo)
              >> needcaseword("NonPads", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numFixedNonPads >> needeol(lineNo);

      PartitionIds freeToMove;
      freeToMove.setToAll(numPart);
      if (_fixedConstr==NULL)
      {
          _fixedConstr=new Partitioning((*_hgraph).getNumNodes(),freeToMove);
      }

      for(k=0;k!=numFixedPads+numFixedNonPads;k++)
      {
         char nodeId[256];
         fixFile >> eathash(lineNo) >> isword(lineNo) >> nodeId 
                 >> noeol(lineNo)   >> needword(":",lineNo) >> noeol(lineNo);

         if (nodeId[0]=='a')
         {
             int nodeNum=atoi(nodeId+1);
             abkfatal2(nodeNum>=0 && nodeNum < (numNodes-numTerminals),
                 " Bad node Id in line ", lineNo);
             PartitionIds constr;

             while ( fixFile.peek() != '\n' )
             {
                char partId[256];
                fixFile >> isword(lineNo) >> partId >> eatblank;
                abkfatal3(partId[0]=='b',
                  " NonPad node can be constrained to regular partitions only ",
                  "in line ", lineNo);
                int partNum=atoi(partId+1);
                abkfatal2(partNum>=0 && partNum < numPart,
                  "Bad partition Id in line ", lineNo);
                constr.addToPart(partNum);
             }
             if (constr.isEmpty()) constr=freeToMove;
             (*_fixedConstr)[nodeNum+numTerminals-1]=constr;
         }
         else if (nodeId[0]=='p' )
         {
             int padNum=atoi(nodeId+1)-1;
             abkfatal2(padNum>=0, " Bad pad Id (must be >1) in line ", lineNo);
             abkfatal2(padNum < numTerminals, 
                       " Bad pad Id (number too big) in line ", lineNo);
             PartitionIds constr;

             while ( fixFile.peek() != '\n' )
             {
                char partId[256];
                fixFile >> isword(lineNo) >> partId >> eatblank;
                if (partId[0]=='b')
                {
                    int partNum=atoi(partId+1);
                    abkfatal2(partNum>=0 && partNum < numPart,
                         "Partition Id out of range in line ", lineNo);
                    constr.addToPart(partNum);
                }
                else if (partId[0]=='p' && partId[1]=='b')
                {
                    int blockNum=atoi(partId+2);
                    abkfatal2(blockNum>=0 && blockNum < numPadBlocks,
                         "Block Id out of range in line ", lineNo); 
                    if ((*_terminalToBlock)[padNum]!=UINT_MAX)
                    {
                      char errMess[256];
          sprintf(errMess," Pad %s is constrained to >1 pad block in line %d\n",
                      nodeId, lineNo);
                      abkfatal(0,errMess);
                    }
                    (*_terminalToBlock)[padNum]=blockNum;
                }
                else
                {   
                   char buf[255];
                   sprintf(buf,".\n Got %s ",partId);
                   abkfatal3(0," Block or partition Id expected in line ",
                             lineNo, buf);
                }
             }
             if (constr.isEmpty()) constr=freeToMove;
             (*_fixedConstr)[padNum]=constr;
         }
         else 
         {
            char buf[255];
            sprintf(buf,"\n Got '%s' ",nodeId);
            abkfatal3(0, "Node Id starting with a or p expected in line "
                       ,lineNo,buf);
         }

         fixFile >> needeol(lineNo); 
         
      }
  return true;
}

bool PartitioningProblem ::readSOL(istream& solFile)
{
      abkfatal(_hgraph!=0," _hgraph is not yet initialized in PartProb");
      abkfatal(solFile," Can't open .sol file for reading");

      unsigned k,numWeights=_hgraph->getNumWeights();
      unsigned numTerminals=_hgraph->getNumTerminals();
      unsigned numNodes=_hgraph->getNumNodes();

      int lineNo=1;
      unsigned numPart=0, numPadBlocks=0, numFixedPads=0, numFixedNonPads=0;

      solFile >> eathash(lineNo);
      solFile >> needcaseword("UCLA",lineNo) 
              >> needcaseword("sol",lineNo) ;
      solFile >> skiptoeol;
      lineNo++;

      solFile >> eathash(lineNo);
      solFile >> needcaseword("Regular",lineNo)
              >> needcaseword("partitions", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numPart >> needeol(lineNo);

      abkfatal(numPart>1,  " Need at least two regular partitions");
      abkfatal(numPart<=32," Can't have more than 32 regular partitions");
      PartitionIds freeToMove, nowhere;
      freeToMove.setToAll(numPart);


      if (!_solnBuffers)
      {
         _solnBuffers=new PartitioningBuffer(_hgraph->getNumNodes(), 1);
         for(unsigned kk=0;kk!=_hgraph->getNumTerminals();kk++)
            (*_solnBuffers)[0][kk]=nowhere;
         for(;k!=_hgraph->getNumNodes();k++) (*_solnBuffers)[0][k]=freeToMove;
      }

      if (_partitions!=NULL)
      {
          abkfatal3((*_partitions).size()==numPart,
              "Mismatch with #partitions from blk file: line ",lineNo,"\n");
      }

      if (_fixedConstr)
      {
          if (_fixedConstr==NULL)
          _fixedConstr=new Partitioning((*_hgraph).getNumNodes(),freeToMove);
      }

      if (_capacities!=NULL)
      {
          abkfatal2((*_capacities).size()==numPart,
              "Mismatch with #partitions from blk file: line ",lineNo);
      }
      else _capacities=new vector< vector<double> > (numPart);

      solFile >> eathash(lineNo) >> needcaseword("Pad",lineNo)
              >> needcaseword("partitions", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numPadBlocks >> needeol(lineNo);

      solFile >> eathash(lineNo) >> needcaseword("Fixed",lineNo)
              >> needcaseword("Pads", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numFixedPads >> needeol(lineNo);
      solFile >> eathash(lineNo) >> needcaseword("Fixed",lineNo)
              >> needcaseword("NonPads", lineNo) >> needword(":",lineNo)
              >> isnumber(lineNo)>> numFixedNonPads >> needeol(lineNo);

      for(k=0;k!=numFixedPads+numFixedNonPads;k++)
      {
         char nodeId[256];
         solFile >> eathash(lineNo) >> isword(lineNo) >> nodeId 
                 >> noeol(lineNo)   >> needword(":",lineNo) >> noeol(lineNo);

         PartitionIds constr;
         if (nodeId[0]=='a')
         {
             int nodeNum=atoi(nodeId+1);
             abkfatal2(nodeNum>=0 && nodeNum < (numNodes-numTerminals),
                 " Bad node Id in line ", lineNo);

             while ( solFile.peek() != '\n' )
             {
                char partId[256];
                solFile >> isword(lineNo) >> partId >> eatblank;
                abkfatal3(partId[0]=='b',
                  " NonPad node can be constrained to regular partitions only ",
                  "in line ", lineNo);
                int partNum=atoi(partId+1);
                abkfatal2(partNum>=0 && partNum < numPart,
                  "Bad partition Id in line ", lineNo);
                constr.addToPart(partNum);
             }
             if (!constr.isEmpty())
                (*_solnBuffers)[0][numTerminals+nodeNum]=constr;
             else
                (*_solnBuffers)[0][numTerminals+nodeNum]=freeToMove;
         }
         else if (nodeId[0]=='p' )
         {
             int padNum=atoi(nodeId+1)-1;
             abkfatal2( (padNum>=0 && padNum < numTerminals),
                 " Bad pad Id in line ", lineNo);

             while ( solFile.peek() != '\n' )
             {
                char partId[256];
                solFile >> isword(lineNo) >> partId >> eatblank;
                if (partId[0]=='b')
                {
                    int partNum=atoi(partId+1);
                    abkfatal2(partNum>=0 && partNum < numPart,
                         "Partition Id out of range in line ", lineNo);
                    constr.addToPart(partNum);
                }
                else if (partId[0]=='p' && partId[1]=='b')
                {
                     // ignore
                }
                else
                {   
                   char buf[255];
                   sprintf(buf,".\n Got %s ",partId);
                   abkfatal3(0," Block or partition Id expected in line ",
                             lineNo, buf);
                }
             }
             if (!constr.isEmpty())
               (*_solnBuffers)[0][padNum]=constr;
             else
               (*_solnBuffers)[0][padNum]=freeToMove;
         }
         else 
         {
            char buf[255];
            sprintf(buf,"\n Got '%s' ",nodeId);
            abkfatal3(0, "Node Id starting with a or p expected in line "
                       ,lineNo,buf);
         }
         solFile >> needeol(lineNo);
      }
      for (k=1; k < _solnBuffers -> size(); k++)
          (*_solnBuffers)[k]=(*_solnBuffers)[0];
  return true;
}

bool PartitioningProblem
::read(const char* netDFileName, const char* areFileName, 
       const char* blkFileName,  const char* fixFileName, 
       const char* solFileName) 
{
   _ownsData = true;
   _costs=vector<double>(1,DBL_MAX);
   _violations=vector<double>(1,DBL_MAX);
   _imbalances=vector<double>(1,DBL_MAX);

   abkfatal(netDFileName,  
       " Can't construct PartitioningProblem: NULL .netD file name passed");

   _hgraph=new HGraphFixed(netDFileName,areFileName);
   cout << " Done reading " << netDFileName;
   if (areFileName) cout << " and " << areFileName;
   cout << endl;
   unsigned numWeights=_hgraph->getNumWeights();

   itHGFNodeGlobalMutable v=_hgraph->nodesBegin();

	//pads no longer have their areas cleared
   //itHGFNodeGlobalMutable z=_hgraph->nodesBegin()+_hgraph->getNumTerminals();
   //for(v; v!=z; v++) 
   //     for(unsigned k=0;k!=numWeights;k++) (*v)->setWeight(0.0,k);

	//count the total weight of the nodes in the problem (including pads)
   _totalWeight = new vector<double>(numWeights, 0.0);
   for (v = _hgraph->nodesBegin(); v != _hgraph->nodesEnd(); v++)  
   {
        for(unsigned k=0; k!=numWeights; k++)
           (*_totalWeight)[k] += (*v)->getWeight(k);
   }

   if (blkFileName)
   {
      cout << " Reading " << blkFileName << " ... " << endl;
      readBLK( ifstream (blkFileName) ); 
   }

   if (fixFileName)
   {
      cout << " Reading " << fixFileName << " ... " << endl;
      readFIX( ifstream (fixFileName) ); 
   }
   else //still need _terminalToBlock
   {
      abkwarn(_hgraph->getNumTerminals()==0,
          "Terminals specified in .net or .netD file, but "
          "no .fix file given.\n  Making all nodes nonterminal");

      //for(unsigned i=0; i!=_hgraph->getNumTerminals();i++)
      //for(unsigned j=0; j!=_hgraph->getNumWeights();  j++)
      //  _hgraph->getNodeByIdx(i).setWeight(0.0,j);

      _hgraph->clearTerminals();	//set num terminals to 0
      abkfatal(_terminalToBlock==0, " _terminalToBlock is already init'd");
      _terminalToBlock=new vector<unsigned>(_hgraph->getNumTerminals(),
                                            UINT_MAX);
   }

   unsigned numPart=_capacities->size();
   PartitionIds freeToMove, nowhere;
   freeToMove.setToAll(numPart);
   if (_capacities)
   {
      _solnBuffers=new PartitioningBuffer(_hgraph->getNumNodes(),1);
     for(unsigned k=0;k!=_hgraph->getNumTerminals();k++)
            (*_solnBuffers)[0][k]=nowhere;
     for(;k!=_hgraph->getNumNodes();k++) (*_solnBuffers)[0][k]=freeToMove;
   }


   if (solFileName)
   {
      cout << " Reading " << solFileName << " ... " << endl;
      readSOL( ifstream (solFileName) );
   }

   if (!_solnBuffers)
   {
      _solnBuffers=new PartitioningBuffer(_hgraph->getNumNodes(),1);
      for(unsigned k=0;k!=_hgraph->getNumTerminals();k++)
            (*_solnBuffers)[0][k]=nowhere;
      for(;k!=_hgraph->getNumNodes();k++) (*_solnBuffers)[0][k]=freeToMove;
   }

   return true;
}
