/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

// Created: 98/02/15 by Andrew Caldwell (caldwell@cs.ucla.edu)

//CHANGES
//980215 aec created capoParams.cxx from capoPlacer.cxx
//980313 ilm added CapoPlacer::Parameters ctor from argc, argv
//           reworked CapoPlacer::Parameters and its op<<
//980325 ilm added n,m and partFuzziness to CapoPlacer::Parameters
//980401 ilm added savePartProb parameter

#include "abkcommon.h"
#include "capoParams.h"

void CapoParameters::setDefaults()
{
    verb=Verbosity("1 1 1");

    stopAtBlocks=0;	
    mlParams.netThreshold = 0;//don't threshold

    plotBlocks        =false;
    saveLayerBBoxes   =false;
    saveBlocks        =false;

    smallPartThreshold  = 35;
    smallPlaceThreshold = 7;

    doOverlapping         = false;
    startOverlappingLayer = 0;
    endOverlappingLayer   = UINT_MAX;

    replaceSmallBlocks = AtTheEnd;
    useActualPinLocs   = false;

    lookAhead          = 0;
    allowRowSplits     = false;
    propFromPl	       = false;
    useKPl	       = false;
}

CapoParameters::CapoParameters()
	: verb("1 1 1")
{  setDefaults(); }

CapoParameters::CapoParameters(const CapoParameters &srcParams)
	:verb(srcParams.verb),
	 stopAtBlocks(srcParams.stopAtBlocks), 
	 replaceSmallBlocks(srcParams.replaceSmallBlocks),
	 useActualPinLocs(srcParams.useActualPinLocs),
	 smallPartThreshold(srcParams.smallPartThreshold),
	 smallPlaceThreshold(srcParams.smallPlaceThreshold),
	 doOverlapping(srcParams.doOverlapping),
	 startOverlappingLayer(srcParams.startOverlappingLayer),
	 endOverlappingLayer(srcParams.endOverlappingLayer),
	 propFromPl(srcParams.propFromPl),
	 useKPl(srcParams.useKPl),
	 mlParams(srcParams.mlParams), 
	 smplParams(srcParams.smplParams),
	 lookAhead(srcParams.lookAhead),
	 allowRowSplits(srcParams.allowRowSplits),
	 splitterParams(srcParams.splitterParams),
         plotBlocks(srcParams.plotBlocks),
         saveLayerBBoxes(srcParams.saveLayerBBoxes),
	 saveBlocks(srcParams.saveBlocks)
{}

CapoParameters::CapoParameters(int argc, const char *argv[])
	:mlParams(argc,argv), smplParams(argc,argv),
	 splitterParams(argc, argv)
{   
  setDefaults();

  BoolParam      helpRequest0   ("help",  argc, argv);
  BoolParam      helpRequest1   ("h",     argc, argv);

  if (helpRequest0.found() || helpRequest1.found() )
  {  printHelp(); return; }

//----------------------------------------------------------------

  StringParam    verb_          ("verb",         argc, argv);

  //general Capo params..
  UnsignedParam  stopAtBlocks_  ("stopAtBlocks", argc, argv);
  UnsignedParam  smallPartThreshold_("smallPartThreshold", argc, argv);
  UnsignedParam  smallPlaceThreshold_("smallPlaceThreshold", argc, argv);
  StringParam    replaceSmallBlocks_("replaceSmallBlocks",   argc, argv);
  BoolParam      useActualPinLocs_("useActualPinLocs", argc,argv);

  //Partitioner params...
  UnsignedParam  lookAhead_      ("lookAhead", argc, argv);
  BoolParam      propFromPl_     ("propFromPl",     argc, argv);              
  BoolParam      useKPl_         ("useKPl",     argc, argv);
  BoolParam      allowRowSplits_ ("allowRowSplits", argc, argv);

  //Optional info-tracking params
  BoolParam	plotBlocks_       ("plotBlocks",       argc, argv);
  BoolParam	saveLayerBBoxes_  ("saveLayerBBoxes",  argc, argv);
  BoolParam     saveBlocks_       ("saveBlocks",       argc, argv);

  //Overlapping Parameters
  BoolParam     doOverlapping_  ("doOverlapping",argc,argv);
  UnsignedParam startOverlappingLayer_("startOverlappingLayer",argc,argv);
  UnsignedParam endOverlappingLayer_  ("endOverlappingLayer",argc,argv);


//----------------------------------------------------------------

  if(smallPartThreshold_.found())
	smallPartThreshold = smallPartThreshold_;
  if(smallPlaceThreshold_.found())
	smallPlaceThreshold = smallPlaceThreshold_;

  if(replaceSmallBlocks_.found())
  {
      if(!strcasecmp(replaceSmallBlocks_,"Never"))
	replaceSmallBlocks = Never;
      else if(!strcasecmp(replaceSmallBlocks_,"AtTheEnd"))
	replaceSmallBlocks = AtTheEnd;
      else if(!strcasecmp(replaceSmallBlocks_,"AtEveryLayer"))
	replaceSmallBlocks = AtEveryLayer;
      else
	abkfatal(0,"invalid string follows '-replaceSmallBlocks'");
  }

  if (useActualPinLocs_.found())
	useActualPinLocs = true;

  if (verb_.found()) verb=Verbosity(verb_);

  if (stopAtBlocks_.found()) stopAtBlocks=stopAtBlocks_;

  if(propFromPl_.found()) propFromPl = true;
  if(useKPl_.found())     useKPl     = true;

  plotBlocks        = plotBlocks_.found();
  saveLayerBBoxes   = saveLayerBBoxes_.found();
  saveBlocks        = saveBlocks_.found();

  if(doOverlapping_.found())
	doOverlapping = true;
  if(startOverlappingLayer_.found())
	startOverlappingLayer = startOverlappingLayer_;
  if(endOverlappingLayer_.found())
      	endOverlappingLayer = endOverlappingLayer_;

  if(lookAhead_.found())
	lookAhead = lookAhead_;

  if(allowRowSplits_.found())
	allowRowSplits = true;
}

ostream& operator<<(ostream& os, CapoParameters &params)
{
    static const char *tfStr[2] = {"false", "true"} ;

    os<<"---- Capo Parameters     "<<endl<<endl;

    os<<" StopAtBlocks:  ";
    if(params.stopAtBlocks)
        os<< params.stopAtBlocks      <<endl;
    else
	cout<<" complete placement"<<endl;

    os<<" SmallPartThreshold:  " << params.smallPartThreshold<<endl;
    os<<" SmallPlaceThreshold: " << params.smallPlaceThreshold<<endl;

    os<<" ReplaceSmallBlocks:  ";
      if(params.replaceSmallBlocks == Never)
	cout<<"Never"<<endl;
      else if(params.replaceSmallBlocks == AtTheEnd)
	cout<<"AtTheEnd"<<endl;
      else if(params.replaceSmallBlocks == AtEveryLayer)
	cout<<"AtEveryLayer"<<endl;
      else 
	abkfatal(0,"invalid value for replaceSmallBlocks");

    os<<" Use Actual Pin Locations: "<<tfStr[params.useActualPinLocs]<<endl;
  
     cout<<" Propigate terminals from globally placed locs: "
	<<tfStr[params.propFromPl]<<endl;

     cout<<" Use (KPlacer)BoxPlacer for splitting: "
	<<tfStr[params.useKPl]<<endl;

    os<<" Allow RowSplits :  "<<tfStr[params.allowRowSplits]<<endl;

    os<<endl<<"----Block Splitting Parameters----"<<endl;
    os<<" LookAhead       :  "<< params.lookAhead<<endl;
    os<<params.splitterParams<<endl;

    os<<endl<<"----Overlapping Parameters----"<<endl;
    os<<" Do Overlapping:      " << tfStr[params.doOverlapping]<<endl;
    if(params.doOverlapping)
    {
	if(params.startOverlappingLayer != 0)
	    os<<" StartOverlappingLayer:    "<<
			params.startOverlappingLayer<<endl;
	if(params.endOverlappingLayer != UINT_MAX)
	    os<<" EndOverlappingLayer:      "<<
			params.endOverlappingLayer<<endl;
    }

    os<<endl<<"----Small Placer Parameters----"<<endl;
    os<<params.smplParams<<endl;

    os<<endl<<"----Data Tracking Parameters----"<<endl<<endl;
    os<<" Plot blocks:         " << tfStr[params.plotBlocks     ] <<endl;
    os<<" Save Layer BBoxes:   " << tfStr[params.saveLayerBBoxes] <<endl;
    os<<" Save Blocks:         " << tfStr[params.saveBlocks     ] <<endl;

    os<<endl<<params.mlParams;

    return os;
}

void CapoParameters::printHelp()
{
     cout << " ******* CAPO PARAMETERS HELP ********* " << endl
          << "== Top-down parameters == \n"
          << "  -verb  Verbosity \n"
          << "  -stopAtBlocks #  \n"
	  << "  -smallPartThreshold  <unsigned> \n"
	  << "  -smallPlaceThreshold <unsigned> \n"
	  << "  -replaceSmallBlocks <Never|AtTheEnd|AtEveryLayer> \n"
	  << "  -useActualPinLocs   \n\n"

	  << "== Block Splitting Parameters =="<<endl
          << "  -noRepart        <bool>         \n"
          << "  -constTol        <double %>[10] \n"
	  << "  -branchingFactor <unsigned> [2] \n"
	  << "  -useNameHier     <delimiting char> [none] \n"
          << "  -lookAhead       <unsigned [0]  \n"
	  << "  -propFromPl      <bool [false]> \n"
	  << "  -useKPl          <bool [false]> \n"
	  << "  -noRowSplits             <bool> \n\n"

	  << "== Optional info-tracking parameters == \n"
	  << "  -plotBlocks        <bool> [false] \n"
          << "  -saveLayerBBoxes   <bool> [false] \n\n"
          << "  -saveBlocks        <bool> [false] \n\n"

	  << "== Overlapping parameters == \n"
          << "  -doOverlapping \n"
          << "  -startOverlappingLayer <unsigned> \n"
	  << "  -endOverlappingLayer <unsigned> \n\n"

	  << "== Small Placer Parameters == \n"
	  << smplParams
	
	  <<endl;
}


// ---------------------------------------------------
// implementation of CapoSplitterParams

CapoSplitterParams::CapoSplitterParams(int argc, const char *argv[])
        : doRepartitioning(true), useWSTolMethod(true),
          constantTolerance(20), numMLSets(2), 
	  useNameHierarchy(false), delimiters("/"), verb(argc,argv)
{
   BoolParam     noRepart_("noRepart", argc, argv);
   UnsignedParam constTol_("constTol", argc, argv);
   UnsignedParam bFactor_("branchingFactor", argc, argv);
   StringParam   useNameHier_("useNameHier", argc, argv);
 
   if(noRepart_.found()) doRepartitioning = false;
 
   if(constTol_.found())
   {
        useWSTolMethod =false;
        constantTolerance = constTol_;
   }
 
   if(bFactor_.found())
        numMLSets = bFactor_;

   if(useNameHier_.found())
   {
	useNameHierarchy = true;
	delimiters = new char[strlen(useNameHier_)+2];
	strcpy(delimiters, useNameHier_);
   }
}
 
ostream& operator<<(ostream& os, const CapoSplitterParams& params)
{
  char* tfStr[2];
  tfStr[0] = "false";
  tfStr[1] = "true";
 
  os<<"---Capo Splitter Parameters---"<<endl;
  os<<"  doRepartitioning   "<<tfStr[params.doRepartitioning]<<endl;
  if(params.useWSTolMethod)
      os<<"  h-cuts use WS Tolerance Method     "<<endl;
  else
      os<<"  h-cuts use constant tolerance "<<params.constantTolerance
		<<"%"<<endl;
  os<<"  v-cuts use constant tolerance "<<params.constantTolerance<<"%"<<endl;

  os<<"  use name hierarchy    "<<tfStr[params.useNameHierarchy]<<endl;
  if(params.useNameHierarchy)
	os<<"   delimiter          "<<params.delimiters<<endl;
 
  return os;
}
                                           

